from .__class_init__ import *
from ..types import choices
from ..types.implementations import structs
threshold_scaler_unit = cdt.ScalUnitType(b'\x02\x02\x0f\x00\x16\x07')


class Thresholds(cdt.Array):
    TYPE = choices.simple_dt


class ActionSet(cdt.Structure):
    """TODO:"""
    action_up: structs.ActionItem
    action_down: structs.ActionItem


class Actions(cdt.Array):
    """Defines the scripts to be executed when the monitored attribute of the referenced object crosses the corresponding threshold. The attribute actions has exactly
    the same number of elements as the attribute thresholds. The ordering of the action_items corresponds to the ordering of the thresholds (see above)."""
    TYPE = ActionSet


class RegisterMonitor(ic.COSEMInterfaceClasses):
    """ DLMS UA 1000-1 Ed.14. 4.5.6. This IC allows modelling the function of monitoring of values modelled by “Data”, “Register”, “Extended register” or “Demand register” objects.
    It allows specifying thresholds, the value monitored, and a set of scripts (see 4.5.2) that are executed when the value monitored crosses a threshold """
    CLASS_ID = ClassID.REGISTER_MONITOR
    VERSION = Version.V0
    A_ELEMENTS = (ic.ICAElement("thresholds", Thresholds),
                  ic.ICAElement("monitored_value", structs.ValueDefinition),
                  ic.ICAElement("actions", Actions))

    def characteristics_init(self):
        ...

    @property
    def thresholds(self) -> Thresholds:
        return self.get_attr(2)

    @property
    def monitored_value(self) -> structs.ValueDefinition:
        return self.get_attr(3)

    @property
    def threshold_normal(self) -> Actions:
        return self.get_attr(4)



import datetime
from .__class_init__ import *


class SpecDayEntry(cdt.Structure):
    """ Specifies a special day identifier for a given date. The date may have wildcards for repeating special days like Christmas. """
    index: cdt.LongUnsigned
    specialday_date: cst.OctetStringDate
    day_id: cdt.Unsigned


class Entries(cdt.Array):
    """ Specifies the list of spec_day_entry """
    TYPE = SpecDayEntry
    values: list[SpecDayEntry]
    __getitem__: SpecDayEntry

    def new_element(self) -> SpecDayEntry:
        indexes: list[int] = [int(el.index) for el in self.values]
        for i in range(0x1_00_00):
            if i not in indexes:
                return SpecDayEntry((i, None, None))  # TODO: insert first DayID from ActiveCalendar as 3 element
        raise ValueError(F'in {self} all indexes is busy')

    def __check_index(self, value):
        """validate day_id from DayProfile"""
        if cdt.LongUnsigned(value) in (entry.index for entry in self.values):
            raise ValueError(F'{cdt.LongUnsigned(value)} already exist in {self}')
        else:
            """validate OK"""

    def validate_exist_index(self, value):
        """pass if value in indexes"""
        if cdt.LongUnsigned(value) not in (entry.index for entry in self.values):
            raise ValueError(F'{cdt.LongUnsigned(value)} not exist in {self}')
        else:
            """validate OK"""

    def get_indexes(self) -> list[int]:
        """ getter for callback Index """
        return [int(entries_element.index) for entries_element in self.values]


class SpecialDaysTable(ic.COSEMInterfaceClasses):
    """ The interface class allows defining dates, which will override normal switching behaviour for special days. The interface class works in
    conjunction with the class "Schedule" or "Activity calendar" and the linking data item is day_id """
    CLASS_ID = ClassID.SPECIAL_DAYS_TABLE
    VERSION = Version.V0
    A_ELEMENTS = ic.ICAElement("entries", Entries),
    M_ELEMENTS = (ic.ICMElement("insert", SpecDayEntry),
                  ic.ICMElement("delete", cdt.LongUnsigned))  # Todo: was Delete.with_cb(None, self.entries.get_indexes)

    def characteristics_init(self):
        self.cardinality = (0, 1)

        self._cbs_attr_post_init.update({2: self.__set_delete})
        self.set_attr(2, None)

    @property
    def entries(self) -> Entries:
        return self.get_attr(2)

    @property
    def insert(self) -> SpecDayEntry:
        return self.get_meth(1)

    @property
    def delete(self) -> cdt.LongUnsigned:
        return self.get_meth(2)

    def __set_delete(self):
        try:
            self.delete.register_cb_preset(self.entries.validate_exist_index)
            self.insert.index.register_cb_preset(self.entries.validate_exist_index)
        except KeyError:  # At init time
            print('set delete NO:')

    def __delete_entry(self):
        """remove one entry by according delete method index. Call after execute"""
        for entry in self.entries.values:
            if entry.index == self.delete:
                self.entries.values.remove(entry)
                return
        else:
            raise ValueError(F'not found entry with index {self.delete} for remove')



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection, overview


class TestType(unittest.TestCase):

    def test_ScalerUnitType(self):
        su = cdt.ScalUnitType((0, 27))
        print(su)

    def test_report(self):
        su = cdt.ScalUnitType((-104, 27))
        rep = su.get_report()
        print(rep)



from ..__class_init__ import *
from ...types.implementations import integers
from . image_transfer_status import ImageTransferStatus


class ImageToActivateInfoElement(cdt.Structure):
    image_to_activate_size: cdt.DoubleLongUnsigned
    image_to_activate_identification: cdt.OctetString
    image_to_activate_signature: cdt.OctetString


class ImageToActivateInfo(cdt.Array):
    """ Provides information on the Image(s) ready for activation. It is generated as the result of the Image verification process. Thi client may check this information before
    activating the Images. """
    TYPE = ImageToActivateInfoElement
    __getitem__: ImageToActivateInfoElement


class ImageTransferInitiate(cdt.Structure):
    """ Initializes the Image transfer process. After a successful invocation of the method the image_transfer_status attribute is set to (1) and the
    image_first_not_transferred_block_number is set to 0. Any subsequent invocation of the method resets the whole Image transfer process and all ImageBlocks need to be
    transferred again. """
    DEFAULT = (bytearray(b'default'), 0)
    image_identifier: cdt.OctetString
    image_size: cdt.DoubleLongUnsigned


class ImageBlockTransfer(cdt.Structure):
    """ Transfers one block of the Image to the server. After a successful invocation of the method the corresponding bit in the image_transferred_block_status attribute
    is set to 1 and the image_first_bit_transferred_block_number attribute is updated """
    image_block_number: cdt.DoubleLongUnsigned
    image_block_value: cdt.OctetString


class ImageTransfer(ic.COSEMInterfaceClasses):
    """ Instance of the Image transfer IC model the process of transferring binary files, called Images to COSEM servers. """
    CLASS_ID = ClassID.IMAGE_TRANSFER
    VERSION = Version.V0
    __blocks_for_update: list[bytearray]
    A_ELEMENTS = (ic.ICAElement("image_block_size", cdt.DoubleLongUnsigned),
                  ic.ICAElement("image_transferred_blocks_status", cdt.BitString, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("image_first_not_transferred_block_number", cdt.DoubleLongUnsigned, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("image_transfer_enabled", cdt.Boolean),
                  ic.ICAElement("image_transfer_status", ImageTransferStatus, classifier=ic.Classifier.DYNAMIC),
                  ic.ICAElement("image_to_activate_info", ImageToActivateInfo, classifier=ic.Classifier.DYNAMIC))
    M_ELEMENTS = (ic.ICMElement("image_transfer_initiate", ImageTransferInitiate),
                  ic.ICMElement("image_block_transfer", ImageBlockTransfer),
                  ic.ICMElement("image_verify", integers.Only0),
                  ic.ICMElement("image_activate", integers.Only0))

    def characteristics_init(self):
        # TODO: compare image_block_size with ServerMaxReceive
        self.__blocks_for_update = list()
        """ container of blocks for transfer """

    @property
    def image_block_size(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(2)

    @property
    def image_transferred_blocks_status(self) -> cdt.BitString:
        return self.get_attr(3)

    @property
    def image_first_not_transferred_block_number(self) -> cdt.DoubleLongUnsigned:
        return self.get_attr(4)

    @property
    def image_transfer_enabled(self) -> cdt.Boolean:
        return self.get_attr(5)

    @property
    def image_transfer_status(self) -> ImageTransferStatus:
        return self.get_attr(6)

    @property
    def image_to_activate_info(self) -> ImageToActivateInfo:
        return self.get_attr(7)

    @property
    def image_transfer_initiate(self) -> ImageTransferInitiate:
        return self.get_meth(1)

    @property
    def image_block_transfer(self) -> ImageBlockTransfer:
        return self.get_meth(2)

    @property
    def image_verify(self) -> integers.Only0:
        return self.get_meth(3)

    @property
    def image_activate(self) -> integers.Only0:
        return self.get_meth(4)

    @property
    def current_block_transfer(self) -> int:
        return int(self.image_block_transfer.image_block_number)

    def get_n_blocks(self) -> int:
        """get sum of blocks"""
        return len(self.__blocks_for_update)

    def set_next_block(self):
        """ set next block transfer """
        try:
            # TODO: USE bitstring of BlockStatus if it exist(length order by image) for pretty next block
            self.set_block_for_transfer(self.current_block_transfer+1)
        except IndexError:
            raise StopIteration

    def set_block_for_transfer(self, index: int):
        """ set block transfer number. IndexError if index more when blocks size """
        if index > (len(self.__blocks_for_update) - 1):
            raise IndexError(F"got block {index=}, expected 0..{len(self.__blocks_for_update)-1}")
        self.image_block_transfer.set((index, self.__blocks_for_update[index]))

    @property
    def is_image_exist(self) -> bool:
        """ show is existing image in container """
        return bool(len(self.__blocks_for_update))

    def clear_image(self):
        """ clear image container """
        self.__blocks_for_update.clear()



""" COMMON Structs """
from abc import ABC, abstractmethod
from ...types import common_data_types as cdt, cosem_service_types as cst
from . import long_unsigneds


class ActionItem(cdt.Structure):
    script_logical_name: cst.LogicalName
    script_selector: cdt.LongUnsigned


class ValueDefinition(cdt.Structure):
    """ Defines an attribute of an object to be monitored. Only attributes with simple data types are allowed. """
    class_id: long_unsigneds.ClassId
    logical_name: cst.LogicalName
    attribute_index: cdt.Integer


class RestrictionByDate(cdt.Structure):
    from_date: cst.OctetStringDate
    to_date: cst.OctetStringDate


class RestrictionByEntry(cdt.Structure):
    from_entry: cdt.DoubleLongUnsigned
    to_entry: cdt.DoubleLongUnsigned


class CaptureObjectDefinition(cdt.Structure):
    """ Capture objects that are assigned to the object instance. Upon a call of the capture (data) method or automatically
    in defined intervals, the selected attributes are copied into the buffer of the profile. """
    DEFAULT = b'\x02\x04\x12\x00\x08\x09\x06\x00\x00\x01\x00\x00\xff\x0f\x02\x12\x00\x00'
    class_id: long_unsigneds.ClassId
    logical_name: cst.LogicalName
    attribute_index: cdt.Integer
    data_index: cdt.LongUnsigned

    def __hash__(self):
        return int.from_bytes(self.logical_name.contents+self.attribute_index.contents+self.data_index.contents, 'big')


class WindowElement(cdt.Structure):
    start_time: cst.OctetStringDateTime
    end_time: cst.OctetStringDateTime


class AccessRight(cdt.Structure):
    """ TODO: maybe more nested description"""
    attribute_access: cdt.Array
    method_access: cdt.Array


class ObjectListElement(cdt.Structure):
    """common for AssociationLN"""
    class_id: long_unsigneds.ClassId
    version: cdt.Unsigned
    logical_name: cst.LogicalName
    access_rights: AccessRight


class UserListEntry(cdt.Structure):
    """user_list_entry"""
    user_id:   cdt.Unsigned
    user_name: cdt.VisibleString



from typing_extensions import deprecated
from dataclasses import dataclass
import numpy as np
from struct import Struct, pack, unpack_from
from typing import Optional, cast, Iterator
import re
from .. import exceptions as exc
from .obis import OBIS

_pattern = re.compile("((?:\d{1,3}\.){5}\d{1,3})(?::(m?\d{1,3}))?")
Index = Struct("?B")


@dataclass(frozen=True)
class Parameter:
    """
     Parameter ::= SEQUENCE
     {
        ln          Cosem-Object-Instance-Id
        descriptor  OPTIONAL
    }

    index       Unsigned8
    nest_index  Unsigned16
    piece       Unsigned8

    descriptor :: = CHOICE
    {
        attribute   [0] IMPLICIT Desc
        method      [1] IMPLICIT Desc
    }

    Desc ::= SEQUENCE
    {
        index
        SEQUENCE (SIZE(255)) OF nest_index OPTIONAL
        piece OPTIONAL
    }
    """
    value: bytes

    def __bytes__(self) -> bytes:
        return self.value

    @classmethod
    def parse(cls, value: str) -> "Parameter":
        """create from string. Only LN, attr/meth type ddd.ddd.ddd.ddd.ddd.ddd:aaa, ex.: 0.0.1.0.0.255 """
        if (res := _pattern.fullmatch(value)) is None:
            raise ValueError(F"in {cls.__name__}.parse got wrong :{value:}")
        else:
            groups = iter(res.groups())
            ret = bytes(map(int, next(groups).split(".")))
            if (a := next(groups)) is not None:
                if a.startswith('m'):
                    a = a[1:]
                    g1 = 256
                else:
                    g1 = 0
                ret += (g1 + int(a)).to_bytes(2)
            return cls(ret)

    def __eq__(self, other) -> bool:
        if isinstance(other, Parameter):
            return cast("bool", self.value == other.value)
        return NotImplemented

    def __lt__(self, other: "Parameter") -> bool:
        """comparing for sort method"""
        if len(self.value) > len(other.value):
            return True
        else:
            return False

    def __str__(self) -> str:
        if (l := len(self.value)) < 6:
            return "No valid"
        elif l == 7:
            return "No valid Index"
        else:
            res = F"{".".join(map(str, self.value[:6]))}"
        if l > 6:
            res += F":{"m" if self.is_method() else ""}{self.i}"
        if l > 8:
            res += F" {"/".join(map(str, self.elements()))}"
        if self.has_piece():
            res += F"p{self.piece}"
        return res

    def validate(self) -> None:
        if (length := len(self.value)) < 6:
            raise exc.DLMSException(F"Parameter got {length=}, expected at least 6")
        if length == 7:
            raise exc.DLMSException(F"Parameter got wrong index")

    @property
    def has_index(self) -> bool:
        return len(self.value) > 6

    @property
    @deprecated("use obis")
    def ln(self) -> bytes:
        """Logical Name"""
        return self.value[:6]

    def is_method(self) -> bool:
        return self.value[6] == 1

    @property
    def i(self) -> int:
        """attribute or method index"""
        return self.value[7]

    def set_i(self, index: int, is_method: bool = False) -> "Parameter":
        val = Index.pack(is_method, index)
        if len(self.value) == 6:
            tmp = self.value + val
        else:
            tmp = bytearray(self.value)
            tmp[6:8] = val
            tmp = bytes(tmp)
        return self.__class__(tmp)

    def append_validate(self) -> None:
        if (l := len(self.value)) < 7:
            raise exc.DLMSException(F"Parameter must has index before")
        elif l % 2 != 0:
            raise exc.DLMSException(F"Can't append to Parameter with piece")

    def append(self, index: int) -> "Parameter":
        """add new sequence(array or struct) index element"""
        self.append_validate()
        return self.__class__(self.value + pack(">H", index))

    def extend(self, *indexes: int) -> "Parameter":
        self.append_validate()
        return self.__class__(self.value + pack(F">{len(indexes)}H", *indexes))

    def pop(self) -> tuple[Optional[int], int, "Parameter"]:
        """
        :return piece, last index and parent Parameter
        ex.: Parameter("0.0.0.0.0.0:2 1/1/1 p3") => (1, Parameter("0.0.0.0.0.0:2 1/1"))
        """
        if self.has_piece():
            return self.value[-1], int.from_bytes(self.value[-3:-1]), self.__class__(self.value[:-3])
        else:
            return None, int.from_bytes(self.value[-2:]), self.__class__(self.value[:-2])

    def set_piece(self, index: int) -> "Parameter":
        """add new sequence(array or struct) index element"""
        if len(self.value) >= 7:
            return self.__class__(self.value + pack("B", index))
        else:
            raise exc.DLMSException(F"Parameter must has index before")

    def has_piece(self) -> bool:
        if (
            (l := len(self.value)) >= 9
            and l % 2 != 0
        ):
            return True
        else:
            return False

    @property
    def piece(self) -> Optional[int]:
        if self.has_piece():
            return self.value[-1]

    def clear_piece(self) -> "Parameter":
        if self.has_piece():
            return self.__class__(self.value[:-1])

    def elements(self, start: int = 0) -> Iterator[int]:
        """return: index elements nested in attribute, started with"""
        for i in range(8 + start, 8 + 2 * self.n_elements, 2):
            res = int.from_bytes(self.value[i:i+2], "big")
            yield res

    def __iter__(self) -> Iterator[int]:
        for it in self.value[:6]:
            yield it
        if self.value[6] == 1:
            yield -self.i
        else:
            yield self.i

    @property
    def last_element(self) -> int:
        """:return last element index"""
        if self.n_elements == 0:
            raise ValueError("Parameter hasn't elements")
        if self.has_piece():
            val = self.value[-3: -1]
        else:
            val = self.value[-2:]
        return int.from_bytes(val, "big")

    @property
    def n_elements(self) -> int:
        """return: amount of elements nested in attribute"""
        return max(0, (len(self.value) - 8) // 2)

    def set(self,
            a: int = None,
            b: int = None,
            c: int = None,
            d: int = None,
            e: int = None,
            f: int = None
            ) -> "Parameter":
        val = bytearray(self.value)
        if a is not None:
            val[0] = a
        if b is not None:
            val[1] = b
        if c is not None:
            val[2] = c
        if d is not None:
            val[3] = d
        if e is not None:
            val[4] = e
        if f is not None:
            val[5] = f
        return self.__class__(bytes(val))

    def __contains__(self, item: "Parameter"):
        return item.value in self.value

    def __getitem__(self, item) -> Optional[int]:
        if self.n_elements > 0:
            return unpack_from(">H", self.value, item*2 + 8)[0]
        else:
            return None

    @property
    def a(self) -> int:
        return self.value[0]

    @property
    def b(self) -> int:
        return self.value[1]

    @property
    def c(self) -> int:
        return self.value[2]

    @property
    def d(self) -> int:
        return self.value[3]

    @property
    def e(self) -> int:
        return self.value[4]

    @property
    def f(self) -> int:
        return self.value[5]

    @property
    def attr(self) -> "Parameter":
        if self.has_index:
            return Parameter(self.value[:8])
        else:
            raise exc.DLMSException(F"Parameter must has index before")

    @property
    def obj(self) -> "Parameter":
        return Parameter(self.value[:6])

    @property
    def obis(self) -> OBIS:
        return OBIS(self.value[:6])


RANGE64 = bytes(range(65))  # Предвычисленный диапазон 0-64


@dataclass(eq=False, frozen=True)
class ParPattern:
    value: bytes
    positions: tuple[int, ...]
    "7 elements , -1 is SKIP, last element is attribute index if positive and method index else negative"

    def __post_init__(self):
        if 6 > len(self.positions) > 7:
            raise ValueError(f"positions must have exactly 6 elements, got {len(self.positions)}")

    @classmethod
    def parse(cls, pattern: str) -> 'LNPattern':
        """Парсинг строк вида:
        - "a.0.(1,2,3).(0-64).0.f" (6 элементов)
        - "a.0.1.0.0.f:2" (6 элементов + индекс через :)
        - "a.0.1.0.0.f:m2" (6 элементов + метод через :)
        """
        # Разделяем основную часть и индекс
        if ':' in pattern:
            main_part, index_part = pattern.rsplit(':', 1)
            is_method = index_part.startswith('m')
            if is_method:
                index_part = index_part[1:]
        else:
            main_part = pattern
            index_part = None

        # Парсим основную часть (6 элементов)
        parts = main_part.split('.', maxsplit=5)
        if len(parts) < 6:
            parts.extend(['f'] * (6 - len(parts)))  # Дополняем до 6 элементов

        value_parts = []
        positions = []
        current_pos = 0

        for i, part in enumerate(parts[:6]):  # Ровно 6 элементов
            if len(part)==1 and ord(part)==97 + i:  # a-f
                if part=='b':
                    value_parts.append(RANGE64)
                    positions.append(current_pos)
                    current_pos += len(RANGE64)
                else:
                    positions.append(-1)  # SKIP
                continue

            try:
                if part.isdigit():  # Простое число
                    num = int(part)
                    if not 0 <= num <= 255:
                        raise ValueError
                    value_parts.append(bytes([num]))
                    positions.append(current_pos)
                    current_pos += 1
                elif part.startswith(('(', '!(')) and part.endswith(')'):
                    elements = cls._parse_group(part[2 if part.startswith('!(') else 1:-1])
                    if part.startswith('!('):
                        elements = set(range(256)) - elements
                    value_parts.append(bytes(elements))
                    positions.append(current_pos)
                    current_pos += len(value_parts[-1])
                else:
                    raise ValueError
            except ValueError:
                raise ValueError(f"Invalid pattern part: {part} at position {i}")

        # Добавляем индекс (7-й элемент)
        if index_part is not None:
            try:
                index_val = int(index_part)
                if not 0 <= index_val <= 255:
                    raise ValueError
                # Упаковываем: старший бит = is_method, остальные 7 бит = значение
                packed = bytes([(0x80 if is_method else 0) | (index_val & 0x7F)])
                value_parts.append(packed)
                positions.append(current_pos)
            except ValueError:
                raise ValueError(f"Invalid index value: {index_part}")
        else:
            positions.append(-1)  # SKIP для 7-го элемента

        return cls(
            value=b''.join(value_parts),
            positions=tuple(positions)
        )

    @staticmethod
    def _parse_group(group_str: str) -> set[int]:
        """Парсинг групп вида '1,2,3' или '10-20'"""
        elements = set()
        for item in group_str.split(','):
            item = item.strip()
            if '-' in item:
                start, end = map(int, item.split('-'))
                elements.update(range(start, end + 1))
            else:
                elements.add(int(item))
        return elements

    # @functools.cache
    def __hash__(self) -> int:
        print("hash")
        hash_parts = []
        for pos in self.positions:
            if pos == -1:
                hash_parts.append(None)
            else:
                hash_parts.append(self.value[pos:pos + 1])
        return hash(tuple(hash_parts))

    def __eq__(self, other: object) -> bool:
        print("eq")
        if not isinstance(other, ParPattern):
            return False

        for i in range(6):
            pos_self = self.positions[i]
            pos_other = other.positions[i]

            # Проверка соответствия схемы позиций
            if (pos_self == -1) != (pos_other == -1):
                return False

            if pos_self != -1:
                # Сравнение значений
                val_self = self.value[pos_self]
                val_other = other.value[pos_other]
                if val_self != val_other:
                    return False
        return True

    def __str__(self) -> str:
        parts = []
        group_chars = {0: 'a', 1: 'b', 2: 'c', 3: 'd', 4: 'e', 5: 'f'}
        for i in range(6):
            pos = self.positions[i]
            if pos == -1:
                parts.append(group_chars[i])
                continue
            if i < 5:
                next_pos = self.positions[i + 1] if self.positions[i + 1] != -1 else len(self.value)
            else:
                next_pos = len(self.value)
            length = next_pos - pos
            val = self.value[pos]
            if (
                i == 1
                and length == 65
                and self.value[pos:pos + 65] == RANGE64
            ):
                parts.append('(0-64)')
                continue
            if length > 1:
                values = list(self.value[pos:next_pos])
                ranges = []
                start = values[0]
                prev = start
                for v in values[1:]:
                    if v != prev + 1:
                        if start == prev:
                            ranges.append(str(start))
                        else:
                            ranges.append(f"{start}-{prev}")
                        start = v
                    prev = v
                if start == prev:
                    ranges.append(str(start))
                else:
                    ranges.append(f"{start}-{prev}")
                if (
                    len(ranges) == 1
                    and '-' in ranges[0]
                ):
                    parts.append(f"({ranges[0]})")
                else:
                    parts.append(f"({','.join(ranges)})")
            else:
                parts.append(str(val))
        return '.'.join(parts)


@dataclass(frozen=True)
class PatternMatcher:
    """Класс для быстрого сопоставления Parameter с ParPattern"""
    patterns: dict['ParPattern', str]
    _masks: list[tuple[tuple[np.ndarray, ...], str]] = None

    def __post_init__(self):
        # Предварительная обработка шаблонов
        masks = []
        for pattern, value in self.patterns.items():
            pattern_masks = []
            for i in range(7):  # Для всех 7 элементов
                if pattern.positions[i]==-1:  # SKIP
                    mask = np.ones(256, dtype=bool)  # Все значения подходят
                else:
                    start = pattern.positions[i]
                    if i < 6:  # Первые 6 элементов (a-f)
                        end = start + 1
                        values = pattern.value[start:end]
                    else:  # 7-й элемент (индекс)
                        end = start + 1
                        values = pattern.value[start:end]
                        # Для методов учитываем старший бит
                        if values[0] & 0x80:
                            values = bytes([values[0] & 0x7F])

                    mask = np.zeros(256, dtype=bool)
                    for v in values:
                        mask[v] = True
                pattern_masks.append(mask)
            masks.append((tuple(pattern_masks), value))
        object.__setattr__(self, '_masks', masks)

    def find_match(self, param: 'Parameter') -> str:
        """Находит первое совпадение параметра с шаблоном"""
        param_values = (
            param.a, param.b, param.c, param.d, param.e, param.f,
            param.i if param.has_index else -1
        )

        for masks, value in self._masks:
            match = True
            for i in range(7):
                val = param_values[i]
                if val==-1:  # Нет значения (для 7-го элемента)
                    if masks[i].any():  # Если в шаблоне требуется конкретное значение
                        match = False
                        break
                elif not masks[i][val]:
                    match = False
                    break
            if match:
                return value
        return None


if __name__ == "__main__":
    # Создаем тестовые шаблоны
    patterns = {
        ParPattern.parse("a.0.(1,2,3).(0-64).0.f:2"): "Pattern1",
        ParPattern.parse("a.0.1.0.0.f:m2"): "Pattern2",
        # ... другие шаблоны
    }

    # Создаем матчер
    matcher = PatternMatcher(patterns)

    # Тестовый параметр
    param = Parameter.parse("0.0.1.0.0.255:2")

    # Поиск совпадения
    result = matcher.find_match(param)
    print(f"Found: {result}")  # Выведет "Pattern1"


from dataclasses import dataclass
from .cosem_interface_classes import Parameter
from typing import Optional, Iterator
from .types import cdt


@dataclass(frozen=True)
class ParValues[T]:
    par:  Parameter
    data: T

    def __getitem__(self, item):
        if item == 0:
            return self.par
        elif item == 1:
            return self.data
        else:
            raise StopIteration

    def __iter__(self) -> Iterator[Parameter | T]:
        return iter((self.par, self.data))

    def __str__(self):
        return F"{self.par} - {self.data}"


@dataclass(frozen=True)
class ParData(ParValues[cdt.CommonDataType]):
    data: Optional[cdt.CommonDataType]



from dataclasses import dataclass
from typing import Protocol, ClassVar
from functools import cached_property
from .parameter import Parameter


@dataclass
class Base(Protocol):
    OBIS: Parameter

    @cached_property
    def LN(self) -> Parameter:
        return self.OBIS.set_i(1)


@dataclass
class Data(Base):
    @cached_property
    def VALUE(self) -> Parameter:
        return self.OBIS.set_i(2)


@dataclass
class Register(Data):
    @cached_property
    def SCALER_UNIT(self) -> Parameter:
        return self.OBIS.set_i(3)



import unittest
from src.DLMS_SPODES.cosem_interface_classes.parameter import Parameter
from src.DLMS_SPODES.cosem_interface_classes import parameters as prs


class TestType(unittest.TestCase):
    def test_one(self):
        My = prs.Data(Parameter.parse("0.0.0.1.0.255"))
        My2 = prs.Data(Parameter.parse("0.0.1.1.0.255"))

        print(My.LN, My2.LN)



from dataclasses import dataclass
from typing import ClassVar
from typing_extensions import deprecated


@deprecated("use <Parameter>")
@dataclass
class AParameter:
    value: ClassVar[bytes]

    @property
    def i(self) -> int:
        return self.value[6]


@dataclass
class Time(AParameter):
    value = b'\x00\x00\x01\x00\x00\xff\x02'





"""
DLMS UA 1000-1 Ed 14
"""
from dataclasses import dataclass
from functools import lru_cache
from typing_extensions import deprecated
from typing import Iterator, Type, TypeAlias, Callable, Any, Self, Literal, Optional, Protocol
from ..types import cdt, ut, cst
from ..relation_to_OBIS import get_name
import logging
from enum import IntEnum
from itertools import count
from .. import exceptions as exc
from .overview import ClassID, Version
from ..settings import settings
from ..config_parser import get_values
from .. import literals


logger = logging.getLogger(__name__)
logger.level = logging.INFO

logger.info(F'Register start')

_n_class = count(0)


class Classifier(IntEnum):
    """ (dyn.) Classifies an attribute that carries a process value, which is updated by the meter itself.
    (static) Classifies an attribute, which is not updated by the meter itself (e.g. configuration data). """
    NOT_SPECIFIC = 0
    STATIC = 1
    DYNAMIC = 2

    def __str__(self):
        return self.name


SelectiveAccessDescriptor: TypeAlias = ut.SelectiveAccessDescriptor  # TODO: make with subclass


@dataclass(frozen=True)
class ICElement:
    NAME: str

    def __str__(self) -> str:
        try:
            return getattr(settings.am_names, self.NAME)
        except AttributeError:
            return self.NAME


@dataclass(frozen=True)
class ICAElement(ICElement):
    DATA_TYPE: Type[cdt.CommonDataType] | ut.CHOICE
    min: int = None
    max: int = None
    default: int = None
    classifier: Classifier = Classifier.STATIC
    selective_access: Type[SelectiveAccessDescriptor] | None = None

    def get_change(self,
                   data_type: Type[cdt.CommonDataType] | ut.CHOICE = None,
                   classifier: Classifier = None) -> Self:
        return ICAElement(
            NAME=self.NAME,
            DATA_TYPE=self.DATA_TYPE if data_type is None else data_type,
            min=self.min,
            max=self.max,
            default=self.default,
            classifier=self.classifier if classifier is None else classifier,
            selective_access=self.selective_access)


@dataclass(frozen=True)
class ICMElement(ICElement):
    DATA_TYPE: Type[cdt.CommonDataType]


_LN_ELEMENT: ICAElement = ICAElement(
    NAME="logical_name",
    DATA_TYPE=cst.LogicalName)
"""" first element for each COSEM Interface Class"""


class ObjectValidationError(exc.DLMSException):
    """use in validation method of COSEMInterfaceClasses"""
    def __init__(self,
                 ln: cst.LogicalName,
                 i: int,
                 message: str):
        Exception.__init__(self, F"for {ln}: {i}. {message}")
        self.ln = ln
        self.i = i


class EmptyAttribute(exc.DLMSException):
    """need read attribute"""
    def __init__(self,
                 ln: cst.LogicalName,
                 i: int):
        Exception.__init__(self, F"empty {ln}: {i}")
        self.ln = ln
        self.i = i


Name: Literal = Literal[
    "Data",
    "Register",
    "Extended register",
    "Demand register",
    "Register activation",
    "Profile generic",
    "Clock",
    "Script table",
    "Schedule",
    "Special days table",
    "Association SN",
    "Association LN",
    "SAP Assignment",
    "Image transfer",
    "IEC local port setup",
    "Activity calendar",
    "Register monitor",
    "Single action schedule",
    "IEC HDLC setup",
    "IEC twisted pair (1) setup",
    "M-BUS slave port setup",
    "Utility tables",
    "Modem configuration",
    "PSTN modem configuration",
    "Auto answer",
    "Auto connect",
    "PSTN Auto dial",
    "Data protection",
    "Push setup",
    "TCP-UDP setup",
    "IPv4 setup",
    "MAC address setup",
    "PPP setup",
    "GPRS modem setup",
    "SMTP setup",
    "GSM diagnostic",
    "IPv6 setup",
    "S-FSK Phy&MAC setup",
    "S-FSK Active initiator",
    "S-FSK MAC synchronization timeouts",
    "S-FSK MAC counters",
    "IEC 61334-4-32 LLC setup",
    "S-FSK IEC 61334-4-32 LLC setup",
    "S-FSK Reporting system list",
    "ISO/IEC 8802-2 LLC Type 1 setup",
    "ISO/IEC 8802-2 LLC Type 2 setup",
    "ISO/IEC 8802-2 LLC Type 3 setup",
    "Register table",
    "Compact data",
    "Status mapping",
    "Security setup",
    "Parameter monitor",
    "Sensor manager",
    "Arbitrator",
    "Disconnect control",
    "Limiter",
    "M-Bus client",
    "Wireless Mode Q channel",
    "M-Bus master port setup",
    "DLMS/COSEM server M-Bus port setup",
    "M-Bus diagnostic",
    "61334-4-32 LLC SSCS setup",
    "PRIME NB OFDM PLC Physical layer counters",
    "PRIME NB OFDM PLC MAC setup",
    "PRIME NB OFDM PLC MAC functional parameters",
    "PRIME NB OFDM PLC MAC counters",
    "PRIME NB OFDM PLC MAC network administration data",
    "PRIME NB OFDM PLC Application identification",
    "G3-PLC MAC layer counters",
    "G3 NB OFDM PLC MAC layer counters",
    "G3-PLC MAC setup",
    "G3 NB OFDM PLC MAC setup",
    "G3-PLC 6LoWPAN adaptation layer setup",
    "G3 NB OFDM PLC 6LoWPAN adaptation layer setup",
    "Wi-SUN setup",
    "Wi-SUN diagnostic",
    "RPL diagnostic",
    "MPL diagnostic",
    "NTP Setup",
    "ZigBee® SAS startup",
    "ZigBee® SAS join",
    "ZigBee® SAS APS fragmentation",
    "ZigBee® network control",
    "ZigBee® tunnel setup",
    "Account",
    "Credit",
    "Charge",
    "Token gateway",
    "Function control",
    "Array manager",
    "Communication port protection",
    "SCHC-LPWAN setup",
    "SCHC-LPWAN diagnostic",
    "LoRaWAN setup",
    "LoRaWAN diagnostic",
    "ISO/IEC14908 Identification",
    "ISO/IEC 14908 Protocol setup",
    "ISO/IEC 14908 protocol status",
    "ISO/IEC 14908 diagnostic",
    "HS-PLC ISO/IEC 12139-1 MAC setup",
    "HS-PLC ISO/IEC 12139-1 CPAS setup",
    "HS-PLC ISO/IEC 12139-1 IP SSAS setup",
    "HS-PLC ISO/IEC 12139-1 HDLC SSAS setup",
    "LTE monitoring"
]
"""Interface class name row from Table 3 – List of interface classes by class_id"""


@lru_cache(150)
def ClassIDVer2Name(class_id: ClassID, ver: Version) -> Name:
    """Table 3 – List of interface classes by class_id"""
    match int(class_id), int(ver):
        case 1, 0: return "Data"
        case 3, 0: return "Register"
        case 4, 0: return "Extended register"
        case 5, 0: return "Demand register"
        case 6, 0: return "Register activation"
        case 7, 0 | 1: return "Profile generic"
        case 8, 0: return "Clock"
        case 9, 0: return "Script table"
        case 10, 0: return "Schedule"
        case 11, 0: return "Special days table"
        case 12, 0 | 1 | 2 | 3 | 4: return "Association SN"
        case 15, 0 | 1 | 2 | 3: return "Association LN"
        case 17, 0: return "SAP Assignment"
        case 18, 0: return "Image transfer"
        case 19, 0 | 1: return "IEC local port setup"
        case 20, 0: return "Activity calendar"
        case 21, 0: return "Register monitor"
        case 22, 0: return "Single action schedule"
        case 23, 0 | 1: return "IEC HDLC setup"
        case 24, 0 | 1: return "IEC twisted pair (1) setup"
        case 25, 0: return "M-BUS slave port setup"
        case 26, 0: return "Utility tables"
        case 27, 1: return "Modem configuration"
        case 27, 0: return "PSTN modem configuration"
        case 28, 0 | 2: return "Auto answer"
        case 29, 1 | 2: return "Auto connect"
        case 29, 0: return "PSTN Auto dial"
        case 30, 0: return "Data protection"
        case 40, 0 | 1 | 2: return "Push setup"
        case 41, 0: return "TCP-UDP setup"
        case 42, 0: return "IPv4 setup"
        case 43, 0: return "MAC address setup"
        case 44, 0: return "PPP setup"
        case 45, 0: return "GPRS modem setup"
        case 46, 0: return "SMTP setup"
        case 47, 0 | 1 | 2: return "GSM diagnostic"
        case 48, 0: return "IPv6 setup"
        case 50, 0 | 1: return "S-FSK Phy&MAC setup"
        case 51, 0: return "S-FSK Active initiator"
        case 52, 0: return "S-FSK MAC synchronization timeouts"
        case 53, 0: return "S-FSK MAC counters"
        case 55, 1: return "IEC 61334-4-32 LLC setup"
        case 55, 0: return "S-FSK IEC 61334-4-32 LLC setup"
        case 56, 0: return "S-FSK Reporting system list"
        case 57, 0: return "ISO/IEC 8802-2 LLC Type 1 setup"
        case 58, 0: return "ISO/IEC 8802-2 LLC Type 2 setup"
        case 59, 0: return "ISO/IEC 8802-2 LLC Type 3 setup"
        case 61, 0: return "Register table"
        case 62, 0 | 1: return "Compact data"
        case 63, 0: return "Status mapping"
        case 64, 0 | 1: return "Security setup"
        case 65, 0 | 1: return "Parameter monitor"
        case 67, 0: return "Sensor manager"
        case 68, 0: return "Arbitrator"
        case 70, 0: return "Disconnect control"
        case 71, 0: return "Limiter"
        case 72, 0 | 1: return "M-Bus client"
        case 73, 0: return "Wireless Mode Q channel"
        case 74, 0: return "M-Bus master port setup"
        case 76, 0: return "DLMS/COSEM server M-Bus port setup"
        case 77, 0: return "M-Bus diagnostic"
        case 80, 0: return "61334-4-32 LLC SSCS setup"
        case 81, 0: return "PRIME NB OFDM PLC Physical layer counters"
        case 82, 0: return "PRIME NB OFDM PLC MAC setup"
        case 83, 0: return "PRIME NB OFDM PLC MAC functional parameters"
        case 84, 0: return "PRIME NB OFDM PLC MAC counters"
        case 85, 0: return "PRIME NB OFDM PLC MAC network administration data"
        case 86, 0: return "PRIME NB OFDM PLC Application identification"
        case 90, 1: return "G3-PLC MAC layer counters"
        case 90, 0: return "G3 NB OFDM PLC MAC layer counters"
        case 91, 1 | 2: return "G3-PLC MAC setup"
        case 91, 0: return "G3 NB OFDM PLC MAC setup"
        case 92, 1 | 2: return "G3-PLC 6LoWPAN adaptation layer setup"
        case 92, 0: return "G3 NB OFDM PLC 6LoWPAN adaptation layer setup"
        case 95, 0: return "Wi-SUN setup"
        case 96, 0: return "Wi-SUN diagnostic"
        case 97, 0: return "RPL diagnostic"
        case 98, 0: return "MPL diagnostic"
        case 100, 0: return "NTP Setup"
        case 101, 0: return "ZigBee® SAS startup"
        case 102, 0: return "ZigBee® SAS join"
        case 103, 0: return "ZigBee® SAS APS fragmentation"
        case 104, 0: return "ZigBee® network control"
        case 104, 0: return "ZigBee® tunnel setup"
        case 111, 0: return "Account"
        case 112, 0: return "Credit"
        case 113, 0: return "Charge"
        case 115, 0: return "Token gateway"
        case 122, 0: return "Function control"
        case 123, 0: return "Array manager"
        case 124, 0: return "Communication port protection"
        case 126, 0: return "SCHC-LPWAN setup"
        case 127, 0: return "SCHC-LPWAN diagnostic"
        case 128, 0: return "LoRaWAN setup"
        case 129, 0: return "LoRaWAN diagnostic"
        case 130, 0: return "ISO/IEC14908 Identification"
        case 131, 0: return "ISO/IEC 14908 Protocol setup"
        case 132, 1: return "ISO/IEC 14908 protocol status"
        case 133, 1: return "ISO/IEC 14908 diagnostic"
        case 140, 0: return "HS-PLC ISO/IEC 12139-1 MAC setup"
        case 141, 0: return "HS-PLC ISO/IEC 12139-1 CPAS setup"
        case 142, 0: return "HS-PLC ISO/IEC 12139-1 IP SSAS setup"
        case 143, 0: return "HS-PLC ISO/IEC 12139-1 HDLC SSAS setup"
        case 151, 0 | 1: return "LTE monitoring"
        case _: raise exc.ITEApplication(F"not find <Interface class name> with: {class_id=}, {ver=}")


class COSEMInterfaceClasses(Protocol):
    CLASS_ID: ClassID
    VERSION: Version | None = None
    """ Identification code of the version of the class. The version of each object is retrieved together with the logical name and the class_id by reading the object_list 
    attribute of an “Association LN” / ”Association SN” object. Within one logical device, all instances of a certain class must be of the same version."""
    A_ELEMENTS: tuple[ICAElement, ...]
    M_ELEMENTS: tuple[ICMElement, ...] = tuple()  # empty if class not has the methods
    cardinality: tuple[int, int | None]
    __attributes: list[cdt.CommonDataType | None]
    __specific_methods: tuple[cdt.CommonDataType, ...] = None
    _cbs_attr_post_init: dict[int, Callable]
    collection: Any | None  # Collection. todo: remove in future

    def __init__(self, logical_name: cst.LogicalName | bytes | str):
        self.collection = None
        # """ TODO: """
        self.cardinality = (0, None)
        """ (min, max). default is (0, None) from 0 to infinity. If min == max then they are value.   
        Specifies the number of instances of the class within a logical device. value The class shall be 
        instantiated exactly “value” times. min...max. The class shall be instantiated at least “min.” times 
        and at most “max.” times. If min. is zero (0) then the class is optional, otherwise (min. > 0) "min." 
        instantiations of the class are mandatory. """

        self.__attributes = [_LN_ELEMENT.DATA_TYPE(logical_name), *[None] * len(self.A_ELEMENTS)]
        """ Attributes container """

        if self.M_ELEMENTS is not None:
            self.__specific_methods = tuple(el.DATA_TYPE() for el in self.M_ELEMENTS)
            """Specific methods container"""

        self._cbs_attr_post_init = dict()
        """container with callbacks for post initial attribute by index"""

        self._cbs_attr_before_init = dict()
        """container with callbacks for before initial attribute by index"""

        # init all attributes with default value
        for i in range(2, len(self.A_ELEMENTS)+2):
            default = self.get_attr_element(i).default
            if default is not None:
                self.set_attr(i, default)

        self.characteristics_init()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.hash_ = next(_n_class)
        # print(cls.__name__)

    @classmethod
    def get_attr_element(cls, i: int) -> ICAElement:
        """return element by order index. Override in each new class"""
        if i == 1:
            return _LN_ELEMENT
        elif i > len(cls.A_ELEMENTS) + 1:
            raise exc.DLMSException(F"got attribute index: {i}, expected 1..{len(cls.A_ELEMENTS) + 1}")
        else:
            return cls.A_ELEMENTS[i - 2]

    @classmethod
    def get_meth_element(cls, i: int) -> ICMElement:
        """ implement in subclasses with methods """
        return cls.M_ELEMENTS[i - 1]

    def characteristics_init(self):
        """ initiate all attributes and methods of class """

    def get_attr(self, index: int) -> Optional[cdt.CommonDataType]:
        if index > (max_l := self.get_attr_length()):
            raise IndexError(F"for {self} got attribute index: {index}, expected 0..{max_l}")
        elif index >= 1:
            return self.__attributes[index-1]
        else:
            raise IndexError(F"not support {index=} as attribute")

    def set_attr_force(self,
                       index: int,
                       value: cdt.CommonDataType):
        self.__attributes[index-1] = value
        """use for change official types to custom(not valid)"""

    def encode(self,
               index: int,
               value: str | int) -> cdt.CommonDataType | None:
        """encode attribute value from string if possible, else return None(for CHOICE variant)"""
        if (attr := self.get_attr(index)) is None:
            data_type = self.get_attr_element(index).DATA_TYPE
            if isinstance(data_type, ut.CHOICE):
                return None
            else:
                return self.get_attr_element(index).DATA_TYPE(value)
        else:
            ret = attr.copy()
            ret.set(value)
            return ret

    def set_attr(self,
                 index: int,
                 value=None,
                 data_type: cdt.CommonDataType = None):
        value = self.get_attr_element(index).default if value is None else value
        data_type = self.get_attr_element(index).DATA_TYPE if data_type is None else data_type
        if self.__attributes[index-1] is None:
            new_value = data_type(value)
            if cb_func := self._cbs_attr_before_init.get(index, None):
                cb_func(new_value)
                self._cbs_attr_before_init.pop(index)
            self.__attributes[index-1] = new_value
            if cb_func := self._cbs_attr_post_init.get(index, None):
                cb_func()
                self._cbs_attr_post_init.pop(index)
            else:
                """without callback post init"""
        else:
            self.__attributes[index-1].set(value)

    def parse_attr(self, index: int, value: cdt.Transcript, data_type: cdt.CommonDataType = None):
        """set attribute value by Transcript"""
        dt = self.get_attr_element(index).DATA_TYPE if data_type is None else data_type
        if hasattr(dt, "TAG"):
            self.__attributes[index - 1] = dt.parse(value)
        else:  # maybe CHOICE
            self.__attributes[index - 1] = self.get_attr(index).parse(value)

    def set_attr_link(self, index: int, link: cdt.CommonDataType):
        # self.__attributes[index - 1] = link  # TODO: without validate now for pass load_objects
        if isinstance(link, self.get_attr_element(index).DATA_TYPE):
            self.__attributes[index-1] = link
        else:
            raise ValueError(F'get wrong link: {link} for {self} attr: {index}')

    def get_attr_data_type(self, index: int) -> Type[cdt.CommonDataType] | ut.CHOICE:
        """search data_type attribute value"""
        value: cdt.CommonDataType = self.get_attr(index)
        if value is not None:
            return value.__class__
        else:
            return self.get_attr_element(index).DATA_TYPE

    def clear_attr(self, i: int):
        """use in template"""
        if i > 1:
            self.__attributes[i-1] = None
        else:
            raise ValueError(F'not support clear {self} attr: {i}')

    @deprecated("use get_meth_element")
    def get_meth(self, index: int) -> Any:
        if index >= 1:
            return self.__specific_methods[index-1]
        else:
            raise IndexError(F'not support {index=} as attribute')

    def get_index_with_attributes(self) -> Iterator[tuple[int, cdt.CommonDataType | None]]:
        """ if by initiation order is True then need override method for concrete class"""
        return iter(zip(range(1, self.get_attr_length()+1), self.__attributes))

    def get_attr_length(self) -> int:
        """common attributes amount"""
        return len(self.A_ELEMENTS)+1

    @property
    def it_index_with_meth(self) -> Iterator[tuple[int, cdt.CommonDataType]]:
        return iter(zip(range(1, 20), self.__specific_methods))

    @property
    def logical_name(self) -> cst.LogicalName:
        """ The logical name is always the first attribute of a class. It identifies the instantiation (COSEM object) of this class.
        The value of the logical_name conforms to OBIS (see IEC 62056-61)"""
        return self.get_attr(1)

    def __lt__(self, other: Self):
        return self.logical_name < other.logical_name

    def __setattr__(self, key, value):
        match key:
            case 'VERSION' | 'CLASS_ID' | 'A_ELEMENTS' | 'M_ELEMENTS' as prop: raise ValueError(F"Don't support set {prop}")
            case _:                                                            super().__setattr__(key, value)

    def __getitem__(self, item) -> cdt.CommonDataType:
        """ get attribute value by index, start with 1 """
        if isinstance(item, str):
            return super(COSEMInterfaceClasses, self).__getattr__(item)
        return self.get_attr(item)

    def __iter__(self) -> Iterator[cdt.CommonDataType]:
        """ return attributes iterator"""
        return iter(self.__attributes)

    def __str__(self):
        return F"{self.logical_name.get_report()} {get_name(self.logical_name)}"

    def get_obis(self) -> bytes:
        """ return obis as bytes[6] """
        return self.logical_name.contents

    @property
    def instance_id(self) -> cdt.OctetString:
        return self.logical_name

    # TODO: rewrite this
    def get_attribute_descriptor(self, index: int) -> bytes:
        """ Cosem-Attribute-Descriptor IS/IEC 62056-53 : 2006, 8.3 Useful types """
        return self.CLASS_ID.contents + self.instance_id.contents + ut.CosemObjectAttributeId(index).contents

    @property
    def string_type_cardinality(self) -> str:
        min_cardinality, max_cardinality = self.cardinality
        if min_cardinality == max_cardinality:
            return str(min_cardinality)
        else:
            max_cardinality = str(max_cardinality) if max_cardinality else 'n'
            return F'{str(min_cardinality)}...{max_cardinality}'

    def reset_attribute(self, index: int):
        """ try set default to value """
        self.set_attr(index, self.get_attr_element(index).default)

    def get_attr_descriptor(self,
                            value: int,
                            with_selection: bool = False) -> ut.CosemAttributeDescriptor:
        """ return AttributeDescriptor without selection """
        return ut.CosemAttributeDescriptor((
            self.CLASS_ID,
            ut.CosemObjectInstanceId(self.logical_name.contents),
            ut.CosemObjectAttributeId(value)))

    def get_meth_descriptor(self, value: str | int) -> ut.CosemMethodDescriptor:
        """ TODO """
        match value:
            case int() as index:
                return ut.CosemMethodDescriptor((ut.CosemClassId(self.CLASS_ID.contents),
                                                 ut.CosemObjectInstanceId(self.logical_name.contents),
                                                 ut.CosemObjectMethodId(index)))

    def __hash__(self):
        return hash(self.logical_name)

    def validate(self):
        """procedure for validate class values"""

    def get_value(self, par: bytes) -> cdt.CommonDataType:
        ret = self.get_attr(par[0])
        for i in par[1:]:
            ret = ret[i]
        return ret

    def get_values(self, par: bytes) -> list[cdt.CommonDataType]:
        ret = [self.get_attr(par[0])]
        for i in par[1:]:
            ret.append(ret[-1][i])
        return ret



import unittest
from src.DLMS_SPODES.types import cdt, cst, ut
from src.DLMS_SPODES.cosem_interface_classes import collection
from src.DLMS_SPODES.cosem_interface_classes import activity_calendar
from src.DLMS_SPODES.cosem_interface_classes.overview import ClassID, Version
from src.DLMS_SPODES import exceptions as exc


class TestType(unittest.TestCase):

    def test_Limiter(self):
        col = collection.get_collection(
            manufacturer=b"KPZ",
            server_type=collection.ParameterValue(
                par=bytes.fromhex("0000600102ff02"),
                value=cdt.OctetString("4d324d5f33").encoding),
            server_ver=collection.ParameterValue(
                par=bytes.fromhex("0000000201ff02"),
                value=cdt.OctetString(bytearray(b"1.4.15")).encoding))


        col.set_manufacturer(b'KPZ')
        col.set_firm_ver(collection.ParameterValue(
                par=bytes.fromhex("0000000201ff02"),
                value=cdt.OctetString(bytearray(b"1.4.0")).encoding))
        col.spec_map = col.get_spec()
        tem = col.add(class_id=ClassID.REGISTER,
                version=Version.V0,
                logical_name=cst.LogicalName.from_obis("0.0.96.9.0.255"))
        tem.set_attr(2, bytes.fromhex('10 00 1b'))
        lim = col.add(class_id=ClassID.LIMITER,
                version=Version.V0,
                logical_name=cst.LogicalName.from_obis("0.0.17.0.5.255"))
        # lim.set_attr(3, bytes.fromhex('11 00 00 00 00'))
        self.assertRaises(exc.EmptyObj, lim.set_attr, 3, bytes.fromhex('11 00 00 00 00'))
        lim.set_attr(2, bytes.fromhex('02 03 12 00 03 09 06 00 00 60 09 00 ff 0f 02'))
        print(col, lim)

    def test_ActivityCalendar(self):
        obj: collection.ActivityCalendar = collection.ActivityCalendar("00 00 0d 00 00 ff")
        obj.day_profile_table_active.set(bytes.fromhex("01 01 02 02 11 00 01 01 02 03 09 04 00 00 ff ff 09 06 00 00 0a 00 64 ff 12 00 01"))
        obj.day_profile_table_active.append((1, [("00:00", "00 00 00 00 00 00", 1)]))
        obj.week_profile_table_active.set(bytes.fromhex("01 01 02 08 09 07 44 65 66 61 75 6c 74 11 01 11 00 11 00 11 00 11 00 11 00 11 00"))
        obj.week_profile_table_active.append()
        # obj.week_profile_table_active.append((bytearray(b'\x00'), 0, 0, 0, 0, 0, 0, 0))
        obj.season_profile_active.set(bytes.fromhex("01 01 02 03 09 07 44 65 66 61 75 6c 74 09 0c ff ff 01 01 ff ff ff ff ff 80 00 00 09 07 44 65 66 61 75 6c 73"))
        obj.season_profile_active.append()
        obj.validate()

    def test_d(self):
        d_p = activity_calendar.DaySchedule([
            ("00:00", "00 00 10 00 64 ff", 1),
            ("02:00", "00 00 10 00 64 ff", 2),
            ("01:00", "00 00 10 00 64 ff", 3),
        ])
        sort = sorted(d_p)
        new = activity_calendar.DaySchedule(sort)
        print(d_p, sort)


from . import ver0
from ...types import choices
from ... import exceptions as exc
from ..__class_init__ import *
from ...types.implementations import structs, arrays
from . import authentication_mechanism_name
from . import abstract


class AccessMode(abstract.AccessMode, elements=tuple(range(7))):
    """Version 0 extension"""
    def is_writable(self) -> bool:
        return True if int(self) in (2, 3, 5, 6) else False

    def is_readable(self) -> bool:
        return True if int(self) in (1, 3, 4, 6) else False


class AccessModeMeth(cdt.Enum, elements=(0, 1, 2)):
    """ Enum of access mode for methods """


# TODO: make as subclass of ver0
class AttributeAccessItem(abstract.AttributeAccessItem):
    """ Implemented attribute and it access . Use in Association LN """
    attribute_id: cdt.Integer
    access_mode: AccessMode
    access_selectors: choices.access_selectors

    def abstract_marker(self):
        ...


class AttributeAccessDescriptor(abstract.AttributeAccessDescriptor):
    """ Array of attribute_access_item """
    TYPE = AttributeAccessItem

    def set_read_access(self, attribute_id: cdt.Integer):
        it: AttributeAccessItem
        for it in self:
            if it.attribute_id == attribute_id:
                it.access_mode.set(1)
                break
        else:
            self.append(AttributeAccessItem((attribute_id, AccessMode.parse("1"), None)))


# TODO: make as subclass of ver0
class MethodAccessItem(cdt.Structure):
    """ Implemented method and it access . Use in Association LN """
    method_id: cdt.Integer
    access_mode: AccessModeMeth


class MethodAccessDescriptor(cdt.Array):
    """ Contain all implemented methods """
    TYPE = MethodAccessItem


class AccessRight(cdt.Structure):
    """ TODO: """
    attribute_access: AttributeAccessDescriptor
    method_access: MethodAccessDescriptor


class ObjectListElement(structs.ObjectListElement, access_rights=AccessRight):
    """"""


class ObjectListType(ver0.ObjectListType):
    TYPE = ObjectListElement

    def is_writable(self, ln: cst.LogicalName, indexes: set[int]) -> bool:
        """ index - DLMS object attribute index.
         True: AccessRight is WriteOnly or ReadAndWrite """
        el: ObjectListElement = next(filter(lambda it: it.logical_name == ln, self), None)
        if el is None:
            raise exc.NoObject(F"not find {ln} in object_list")
        item: AttributeAccessItem
        for index in indexes:
            for item in el.access_rights.attribute_access:
                if int(item.attribute_id) == index:
                    if int(item.access_mode) not in (2, 3, 5, 6):
                        return False
                    else:
                        break
                else:
                    continue
            else:
                raise ValueError(F"not find in {ln} attribute index: {index}")
        return True


class ContextNameType(cdt.AXDR, ver0.ApplicationContextName):
    """ In the COSEM environment, it is intended that an application context pre-exists and is referenced by its name during the establishment of an
    application association. This attribute contains the name of the application context for that association."""
    DEFAULT = b'\x09\x07\x60\x85\x74\x05\x08\x01\x01'


class MechanismNameType(cdt.AXDR, authentication_mechanism_name.AuthenticationMechanismName):
    """ In the COSEM environment, it is intended that an application context pre-exists and is referenced by its name during the establishment of an
    application association. This attribute contains the name of the application context for that association."""
    DEFAULT = b'\x09\x07\x60\x85\x74\x05\x08\x02\x00'


class AssociationLN(ver0.AssociationLN):
    """5.4.6 Association LN"""
    VERSION = Version.V1
    A_ELEMENTS = (ic.ICAElement("object_list", ObjectListType, selective_access=ver0.SelectiveAccessDescriptor),
                  ver0.AssociationLN.get_attr_element(3),  # associated_partners_id
                  ic.ICAElement("application_context_name", ContextNameType),
                  ver0.AssociationLN.get_attr_element(5),  # xDLMS_context_info
                  ic.ICAElement("authentication_mechanism_name", MechanismNameType),
                  ic.ICAElement("secret", ver0.LLCSecret),  # TODO: make new class Secret(LLC_Secret)
                  ver0.AssociationLN.get_attr_element(8),  # association_status
                  ic.ICAElement("security_setup_reference", cst.LogicalName))
    M_ELEMENTS = (ver0.AssociationLN.get_meth_element(1),
                  ver0.AssociationLN.get_meth_element(2),
                  ic.ICMElement("add_object", ObjectListElement),
                  ic.ICMElement("remove_object", ObjectListElement))
    object_list: ObjectListType
    application_context_name: ContextNameType
    authentication_mechanism_name: MechanismNameType
    security_setup_reference: cst.LogicalName
    add_object: ObjectListElement
    remove_object: ObjectListElement

    def characteristics_init(self):
        super(AssociationLN, self).characteristics_init()
        # References a "Security setup" object by its logical name. The referenced object manages security for a given "Association LN" object
        # instance.
        self.set_attr(9, None)
